module ObicneDJ
	integer,private:: Nequa	   ! broj jednadzbi
	! prostor za privremene vrijednosti F desne strane:
	real,private,allocatable:: fValue(:),tmpfValue(:)

	!--------- deklaracija varijabli i podataka za RK postupak
	real,private:: CK(4)
	data CK/0.,0.5,0.5,1./
	real,private:: DK(4)
	data DK/0.,0.5,0.5,1./
	real,private:: CY(4)
	data CY/1.,2.,2.,1./
	! prostor za privremene vrijednosti Runge-Kutta koeficijenata:
	real,private,allocatable:: k_RK(:,:)

	!--------- deklaracija varijabli i podataka za RKF postupak
	! prostor za privremene vrijednosti Y varijabli i RKF keof.:
	real,private,allocatable:: tmpYvar(:),k_RKF(:,:)
	real,private:: c_rk(6)     ! rezervacija prostora za koeficijente
	real,private:: d_rk(6,5)   ! reda 4 i 5
	real,private:: a_rk(6)
	real,private:: e_rk(6)

	data (a_rk(i),i=1,6)/ &
	   .1157407,   .0000000,   .5489279,   .5353314,  -.2000000,   .0000000 /
	data (c_rk(i),i=1,6)/ &
	   .0000000,   .2500000,   .3750000,   .9230769,  1.0000000,   .5000000 /
	data (e_rk(i),i=1,6)/ &
	   .1185185,   .0000000,   .5189863,   .5061315,  -.1800000,   .0363636 /
	data ((d_rk(i,j),j=1,5),i=1,6)/ &
	   .0000000,   .0000000,   .0000000,   .0000000,   .0000000, &
	   .2500000,   .0000000,   .0000000,   .0000000,   .0000000, &
	   .0937500,   .2812500,   .0000000,   .0000000,   .0000000, &
	   .8793810, -3.2771960,  3.3208920,   .0000000,   .0000000, &
	  2.0324080, -8.0000000,  7.1734890,  -.2058967,   .0000000, &
	  -.2962963,  2.0000000, -1.3816760,   .4529727,  -.2750000/
	!----------kraj RKF
contains

	logical function OpenObicneDJ(Nequations)
	implicit none
	integer Nequations
	integer ierr
		! alociranje memorije za svaku od n jednadzbi
		! ( memorija se alocira za sve metode, bez obzira ako se ne koriste)
		Nequa = Nequations
		allocate(tmpYvar(Nequa),k_RK(Nequa,4),k_RKF(Nequa,6),fValue(Nequa),tmpfValue(Nequa),stat=ierr)
		! ocisti memorijsko'smece'
		if(allocated(tmpYvar)) tmpYvar=0
		if(allocated(k_RK)) k_RK=0
		if(allocated(k_RKF)) k_RKF=0
		if(allocated(fValue)) fValue=0
		if(allocated(tmpfValue)) tmpfValue=0
		OpenObicneDJ = (ierr==0)	! vraca vrijednost .true. ako je alokacija OK
	end function OpenObicneDJ

	subroutine CloseObicneDJ
	implicit none
		! otpustanje memorije
		if(allocated(tmpYvar)) deallocate(tmpYvar)
		if(allocated(k_RK)) deallocate(k_RK)
		if(allocated(k_RKF)) deallocate(k_RKF)
		if(allocated(fValue)) deallocate(fValue)
		if(allocated(tmpfValue)) deallocate(tmpfValue)
	end subroutine CloseObicneDJ

	!---------- Explicitni postupak
	!  x = x + h
	!  y = y + F(x,y)*h
	logical function Ekspicitni_Korak(Xvar,Yvar,Hstep,FuncValue)
	implicit none
	real Xvar,Yvar(*),Hstep
	external FuncValue
	integer ierr
	integer k
		call FuncValue(Nequa,Xvar,Yvar,fValue,ierr)
		if(ierr.ne.0) then
			Ekspicitni_Korak = .false.
		else
			Xvar=Xvar+Hstep
			do k=1,Nequa
				Yvar(k)=Yvar(k)+fValue(k)*Hstep
			enddo
			Ekspicitni_Korak = .true.
		endif
	end function Ekspicitni_Korak

	!---------- Mjesoviti postupak
	!  x = x + h
	!  y = y + (1-th)*F(x,y)*h + th*F(x+h,y+dy)*h
	logical function Mjesoviti_Korak(Xvar,Yvar,Hstep,Theta,nIter,FuncValue)
	implicit none
	real Xvar,Yvar(*),Hstep,Theta
	integer nIter
	external FuncValue
	integer ierr
	real,automatic:: tmpXvar
	integer k,iter
		! spremanje varijabli s pocetka intervala
		tmpXvar=Xvar
		do k=1,nequa
			tmpYvar(k)=Yvar(k)
		end do
		! desna strana na pocetku intervala: F(x,y)
		call FuncValue(Nequa,tmpXvar,tmpYvar,tmpfValue,ierr)
		if(ierr.ne.0) then
			Mjesoviti_Korak = .false.
		else
			tmpXvar=Xvar+Hstep
			! iteracija za stanje na kraju intervala: F(x+h,y+dy)
			do iter=1,nIter
				call FuncValue(Nequa,tmpXvar,Yvar,fValue,ierr)
				if(ierr.ne.0) then
					Mjesoviti_Korak = .false.
					return
				else
					! popravak stanja na kraju intervala
					do k=1,Nequa
						Yvar(k)=tmpYvar(k)+((1-Theta)*tmpfValue(k)+Theta*fValue(k))*Hstep
					enddo
				endif
			enddo
			Xvar=Xvar+Hstep
			Mjesoviti_Korak = .true.
		endif
	end function Mjesoviti_Korak

	!---------- Runge-Kutta postupak
	! Runge-Kutta metoda 4 reda:
	! k1=h*f(x,y)
	! k2=h*f(f(x+h/2,y+k1/2)
	! k3=h*f(f(x+h/2,y+k2/2)
	! k4=h*f(f(x+h,y+k3)
	! dy=(k1+2k2*2k3+k4)/6
	!  x = x + h
	!  y = y + dy
	logical function RK_Korak(Xvar,Yvar,Hstep,FuncValue)
	implicit none
	real Xvar,Yvar(*),Hstep
	external FuncValue
	real,automatic:: tmpXvar
	integer,automatic:: i,k
	integer ierr
		RK_Korak=.true.
		! izracunavanje Runge-Kutta koeficijenata:
		do k=1,4
			tmpXvar=Xvar+ck(k)*Hstep    ! x+Ck*h
			do i=1,Nequa
				if (k.eq.1) then
					tmpYvar(i)=Yvar(i)
				else
					tmpYvar(i)=Yvar(i)+DK(k)*k_RK(i,k-1)  ! Y+Dk*rK
				endif
			enddo
			call FuncValue(Nequa,tmpXvar,tmpYvar,fValue,ierr)
			if(ierr .ne. 0) then
				RK_Korak = .false.
				return
			else
				do i=1,Nequa
					k_RK(i,k)=fValue(i)*Hstep   ! h*F(x+Ck*h,Yi+Dkj*Kj)
				enddo
			endif
		enddo
		! izracunavanje varijabli na kraju koraka h
		Xvar=Xvar+Hstep
		do i=1,Nequa
			do k=1,4
				Yvar(i)=Yvar(i)+CY(k)*k_RK(i,k)/6.0
			enddo
		enddo
	end function RK_Korak

	!---------- Runge-Kutta-Fehlberg postupak
	!
	!                         Opci poprogram
	!  za integriranje sustava obicnih diferecijalnih jednadzbi
	!
	!                       Y'(x)=F(x,Y)
	! za pocetne uvjete:
	!                        x=xo, Y=Yo
	! u jednom koraku integracije h
	! gdje, su Y,Y' i F vektori:
	!   Y=[y1,y2,y3,...yn], Y'=[y1',y2',y3',...yn'], F=[f1,f2,f3,...fn] 
	! po metodi:
	!
	!               RungeKuttaFehlberg
	!
	! 4 reda sa ocjenom tocnosti
	!
	! Ulaz:
	!       Nequa - broj jednadzbi n
	!       Hstep - korak nezavisne varijable h
	!       FuncValue - externi potprogram za izracunavanje
	!                   desne strane, vidi dolje
	! Ulaz/Izlaz:
	!       Xvar - vrijednost nezavisne varijable
	!              na ulazu vrijednost sa pocetka koraka
	!            - na izlazu vraca vrijednost na kraju koraka
	!       Yvar - vrijednost zavisne varijable, vektor
	!              na ulazu vrijednost sa pocetka koraka
	!            - na izlazu vraca vrijednost na kraju koraka
	! Izlaz:
	!       Eval - vektor apsolutnih vrijednosti pogreske
	!              po odgovarajucim zavisnim varijablama
	! 
	! FuncValue:
	!   je korisnicki podprogram za izracunavanje vektora desne strane:
	!                       F=[f1,f2,f3,...fn]
	! za zadane x i Y, koji mora imati oblik:
	!
	!        subroutine FuncValue(n,x,Y,F,ierr)
	!        integer n  ! broj jednadzbi,ulaz
	!        real x     ! vrijednost nezavisne var.,ulaz
	!        real Y(n)  ! vektor vrijednosti zavisne varijable,ulaz
	!        real F(n)  ! vektor vrijednosti desne strane, izlaz
	!        integer ierr ! indikator greske, ako natupi greska, postaviti
	!                     ! na nenultu vrijednost
	!
	!        ... korisnicki kod za izracunavanje:
	!          F(1) = ...
	!          F(2) = ...
	!          ...
	!          F(n) = ...
	!          ierr=0 !(ako je sve O.K.)
	!        end subroutine
	!   ierr - indikator greske, koju vraca korisnicki podprogram,
	!	razlicito od 0 postoji pogrijeska. Tada	RKF_Step prekida operacije
	!	i vraca vrijednost .false.
	!
	logical function RKF_Step(Xvar,Yvar,Hstep,Eval,FuncValue)
	implicit none
	external FuncValue
	real Xvar,Yvar(*),Hstep,Eval(*)
	real,automatic:: tmpXvar
	integer,automatic:: i,j,k
	integer ierr

	 ! izracunavanje Runge-Kutta koeficijenata:
	 !
	 !     Kik = h*F(x+Ck*h,Yi+Dkj*Kj) ;   i=1,2,3,..n; k=1,2,..6; j=1,..k-1
	 !
	 RKF_Step = .true.
	 do k=1,6
	  tmpXvar=Xvar+c_rk(k)*Hstep    ! x+Ck*h
	  do i=1,Nequa
	   tmpYvar(i)=Yvar(i)
	  enddo
	  do j=1,k-1
	   do i=1,Nequa
		tmpYvar(i)=tmpYvar(i)+d_rk(k,j)*k_RKF(i,j)  ! Yi+Dkj*Kj
	   enddo
	  enddo

	  call FuncValue(Nequa,tmpXvar,tmpYvar,fValue,ierr)
	  if(ierr .ne. 0) then
		RKF_Step = .false.
		return
	  endif

	  do i=1,Nequa
	   if(ierr==0) then
		k_RKF(i,k)=fValue(i)*Hstep   ! h*F(x+Ck*h,Yi+Dkj*Kj)
	   else
	    return
	   endif
	  enddo

	 enddo

	 ! izracunavanje varijabli na kraju koraka h za 4 red
	 ! Y4i = Yi + Ej*Kij ;   i=1,2,3,..n; k=1,2,..6;
	 Xvar=Xvar+Hstep
	 do i=1,Nequa
	  do k=1,6
	   Yvar(i)=Yvar(i)+a_rk(k)*k_RKF(i,k)
	  enddo
	 enddo
	 ! izracunavanje pogreske Abs(Y4-Y5)
	 do i=1,Nequa
	  Eval(i)=0
	  do k=1,6
	   Eval(i)=Eval(i)+(e_rk(k)-a_rk(k))*k_RKF(i,k)
	  enddo
	  Eval(i)=abs(Eval(i))
	 enddo

	end function RKF_Step

	! izracunavanje tolerancije za real X
	! na zadani broj znacajnih znamenaka
	real function Toleranca(x,nZnamenaka)
	implicit none
	real x
	integer nZnamenaka
		if (x.eq.0) then
			Toleranca=5*10**(-nZnamenaka-1)
		elseif (abs(x).gt.1) then
			Toleranca=5 * 10**(aint(Log10(Abs(x)))-nZnamenaka)
		elseif (abs(x).lt.1) then
			Toleranca=5 * 10**(aint(Log10(Abs(x)))-nZnamenaka-1)
		endif
	end function Toleranca


end module ObicneDJ
